<?php

/**
 * @file
 * Drush integration providing common maintenance tasks.
 */

use Faker\Factory;
use Drush\Commands\DrushCommands;
use Drush\Log\LogLevel;

/**
 * Contains DrushDeidentify class.
 */
class DrushDeidentify extends DrushCommands {
  public $sqlCommands;
  public $database;
  public $faker;

  /**
   * Construction.
   */
  public function __construct() {
    // Drush 8.4 ish is strange..
    $sql = drush_sql_get_class();
    // Sqlmysql.
    $db_spec = $sql->db_spec();
    $this->sql = drush_sql_get_class($db_spec);

    $this->database = \Database::getConnection();
    $mpath = drupal_get_path('module', 'drush_deidentify');
    require_once $mpath . '/src/autoload.php';
    $this->faker = Factory::create();
  }

  /**
   * Helper to log messages to the console.
   */
  public function log($message, $status = NULL) {
    // 'warning', 'error', 'success' and 'notice'
    if (!empty($status)) {
      if ($status == 'error') {
        return drush_set_error($message);
      }
    }
    drush_log($message, 'ok');
    // Return drush_print($message);
  }

  /**
   * Exit.
   */
  public function exitClean() {
    die('Command Exit');
  }

  /**
   * Attempt to find primekey for table.
   *
   * @param  array $details
   *   The result of all DESCRIBE table.
   * @return false|mixed
   */
  public function findPrimeKeyForTable(array $details, $table_name) {
    $prime_key_field = FALSE;
    foreach ($details as $col) {
      if (!empty($col['Key'])) {
        if ($col['Key'] == 'PRI') {
          if (!in_array(
            $col['Field'], [
              'deleted',
              'delta',
              'property',
              // 'name'
            ]
          )
          ) {
            if ($col['Type']) {
              // Not using str_contains yet.
              if (strpos($col['Type'], 'int') !== FALSE || strpos($col['Type'], 'varchar') !== FALSE) {
                $prime_key_field = $col['Field'];
              }
            }
          }
        }
      }
    }

    // Here ensure you can alter this field if required. like ($prime_key_field = nid) ect.
    drupal_alter('drush_deidentify_data_clean_force_prime_key', $prime_key_field, $details, $table_name);
    // $this->moduleHandler->invokeAll('drush_deidentify_data_clean_force_prime_key', [&$prime_key_field,  $details, $table_name]);
    return $prime_key_field;
  }

  /**
   * Sets the value dependent on type.
   *
   * @param $type
   * @param $field_name
   * @param $table_name
   * @param $colum_details
   * @param $current_db_item
   *
   * @return int|string
   */
  public function generateFakeValueForField($type, $field_name, $table_name, $colum_details, $current_db_item) {
    $value = '';

    switch ($type) {
      case 'varchar(255)':
        $value = $this->faker->lexify();
        break;

      case 'longtext':
        // SET value here.
        $num = mt_rand(150, 550);
        $value = $this->faker->sentence($num);
        break;

      case 'decimal(14,3)':
      case 'int(11)':
        $value = $this->faker->unixTime();
        break;

      case 'text':
      case 'mediumtext':
        $value = $this->faker->word();
        break;

      default;
        break;
    }

    // Hook here to make granular decisions.
    $context = [
      'type' => $type,
      'field_name' => $field_name,
      'table' => $table_name,
      'col' => $colum_details,
      'faker' => $this->faker,
      'current_data' => $current_db_item,
    ];

    drupal_alter('drush_deidentify_data_clean_colum_val', $value, $context);
    // $this->moduleHandler->invokeAll('drush_deidentify_data_clean_colum_val', [&$value, $type, $field_name, $table_name, $colum_details, $this->faker, $current_db_item]);
    return $value;
  }

  /**
   * Ignore files as an array.
   *
   * @return array
   *   Returns an array.
   */
  public function ignoreFieldNames() {
    return [
      'revision_id',
      'entity_id',
      'deleted',
      'bundle',
      'langcode',
      'delta',
    ];
  }
}

/**
 * Implements hook_drush_command().
 */
function drush_deidentify_drush_command() {
  return [
    // Define the 'db-export' command.
    'drush_deidentify-db-export' => [
      'description' => 'Drush Export Database excluding tables.',
      'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_DATABASE,
      'aliases' => ['db-export'],
      'examples' => ['drush db-export --gzip --result-file=some_db_export_file_name.sql'],
      'options' => [
        'result-file' => '--result-file result file relative to drupal root',
        'gzip' => '--gzip',
        'omit_tables' => 'A comma-separated list of tables to include for , but not data.',
      ],
    ],
    'drush_deidentify-db-clean' => [
      'description' => 'Database Clean after import.',
      'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_DATABASE,
      'aliases' => array('db-clean'),
      'examples' => array('drush db-clean watchdog~message'),
      'options',
    ],
  ];
}

/**
 * Contains drush_drush_deidentify_db_export.
 */
function drush_drush_deidentify_db_export() {
  $ice = new DrushDeidentify();
  $filename = drush_get_option('result-file');
  if (empty($filename)) {
    return $ice->log(dt('Missing output file name example: --result-file=dbexport.sql'), 'error');
  }

  $file_suffix = '';
  $table_selection = $ice->sql->get_expanded_table_selection();
  $default = [
    'cache',
    'cache_admin_menu',
    'cache_block',
    'cache_bootstrap',
    'cache_entity_newsletter',
    'cache_entity_subscriber',
    'cache_features',
    'cache_field',
    'cache_filter',
    'cache_form',
    'cache_hacked',
    'cache_image',
    'cache_libraries',
    'cache_media_oembed',
    'cache_menu',
    'cache_metatag',
    'cache_page',
    'cache_pagepreview',
    'cache_path',
    'cache_rules',
    'cache_token',
    'cache_ultimate_cron',
    'cache_update',
    'cache_views',
    'cache_views_data',
  ];

  $table_selection['structure'] = array_merge($table_selection['structure'], $default);

  if (!empty(drush_get_option('omit_tables'))) {
    $tables = explode(',', drush_get_option('omit_tables'));
    if (!empty($tables) && is_array($tables)) {
      $table_selection['structure'] = array_merge($table_selection['structure'], $tables);
    }
  }

  // Offer hook so no need to alter this module.
  drupal_alter('drush_deidentify_export', $table_selection);

  $ice->log(dt('Db tables structure only  @skiped', ['@skiped' => implode(',', $table_selection['structure'])]));

  $file = $ice->sql->dumpFile($filename);
  $cmd = $ice->sql->dumpCmd($table_selection);
  // Gzip the output from dump command(s) if requested.
  if (drush_get_option('gzip')) {
    $cmd .= ' | gzip -f';
    $file_suffix .= '.gz';
  }
  if ($file) {
    $file .= $file_suffix;
    $cmd .= ' > ' . drush_escapeshellarg($file);
  }

  // Avoid the php memory of the $output array in drush_shell_exec().
  if (!$return = drush_op_system($cmd)) {
    if ($file) {
      drush_log(dt('Database dump saved to !path', array('!path' => $file)), LogLevel::SUCCESS);
      drush_backend_set_result($file);
    }
  }
  else {
    return drush_set_error('DRUSH_SQL_DUMP_FAIL', 'Database dump failed');
  }
}

/**
 * contains drush_drush_deidentify_db_clean.
 */
function drush_drush_deidentify_db_clean($clean_tables = '') {
  $ice = new DrushDeidentify();

  $tables = [];
  if (!empty($clean_tables)) {
    $tables_to_validate = explode(',', $clean_tables);
    foreach ($tables_to_validate as $item) {
      if (strpos($item, '~') !== FALSE) {
        // User Supplied Idea to have per colum
        // "watchdog~message".
        $broken_string = explode('~', $item);
        if (count($broken_string) == 2) {
          $tname = $broken_string[0];
          $cname = $broken_string[1];
          $tables[$tname][] = $cname;
        }
        else {
          return $ice->log(
            dt(
              'Please check the argument @table it does not have required structure eg table~column ', [
                '@table' => $item,
              ]
            ), 'error'
          );

        }
      }
      else {
        return $ice->log(
          dt(
            'Please check the argument @table it does not have required structure eg table~column ', [
              '@table' => $item,
            ]
          ), 'error'
        );
      }
    }
  }

  drupal_alter('drush_deidentify_data_clean', $tables);

  drush_print(dt('Table(s) to clean:  @clean ', ['@clean' => implode(',', array_keys($tables))]));
  if (!drush_confirm('Do you really want to continue?')) {
    //drush_user_abort();
    return;
  }

  $table_cols_values = [];
  $table_prime_keys = [];
  foreach ($tables as $table_name => $columns) {
    try {
      /**
       * @todo find out why the normal...
       * $this->database->query('DESCRIBE {@table}', ['@table' => $table]);
       * does not work ????????
       */
      $query_string = "DESCRIBE `" . trim($table_name) . '`';
      $query = $ice->database->query($query_string);

      $rows = [];
      while ($result = $query->fetchAssoc()) {
        $rows[] = $result;
        if (in_array($result['Field'], $columns)) {
          $table_cols_values[$table_name][] = $result;
          $s = $result['Field'];
          $columns = array_flip($columns);
          unset($columns[$s]);
          if (!empty($columns)) {
            $columns = array_flip($columns);
          }
        }
      }

      if (!empty($columns)) {
        // Computer says no.
        return $ice->log(
          dt(
            'Error: these columns are not found in table -> @key  @cols', [
              '@cols' => json_encode($columns),
              '@key' => $table_name,
            ]
          ), 'error'
        );

      }

      $prime_key = $ice->findPrimeKeyForTable($rows, $table_name);

      if (!empty($prime_key)) {
        if ($prime_key) {
          $ice->log(
            dt(
              'Prime key for @table set to: @key ', [
                '@table' => $table_name,
                '@key' => $prime_key,
              ]
            )
          );
          $table_prime_keys[$table_name] = $prime_key;
        }
      }
      else {
        return $ice->log(dt('Sorry we can\'t find a prime key for @table you will need to use hook_drush_deidentify_data_clean_force_prime_key', ['@table' => $table_name]), 'error');
      }
      // @todo should we die here.
    }
    catch (\Exception $e) {
      return $ice->log(
        dt(
          'Error with argument @table : got @message', [
            '@table' => $table_name,
            '@message' => $e->getMessage(),
          ]
        ), 'error'
      );
    }
  }

  if (empty($tables)) {
    return $ice->log(dt('Think you need to add some tables as an argument eg drush_deidentify:db-clean watchdog~message or by hook_drush_deidentify_data_clean'), 'error');
  }

  if (!drush_confirm('Please confirm above table prime key(s) if not correct stop here.')) {
    $ice->log(dt('use hook_drush_deidentify_data_clean_force_prime_key to set the correct prime.'));
    //drush_user_abort();
    return;
  }

  $ice->log(dt('Cleaning...'));

  if (!empty($table_cols_values)) {
    $field_values = [];
    foreach ($table_cols_values as $table_name => $cols_value) {
      $sql = "SELECT * FROM {" . $table_name . "}";
      $query = $ice->database->query($sql);
      // Here we have all the things.
      $result = $query->fetchAll();
      $num_results = count($result);
      if ($num_results = 0) {
        // No were not generating new stuff thats devel_gennerate job.
        continue;
      }
      else {

        foreach ($result as $item) {
          $fields_and_values = [];
          foreach ($cols_value as $details) {
            $col_name = $details['Field'];
            if (!in_array($col_name, $ice->ignoreFieldNames())) {
              $fields_and_values[$col_name] = $ice->generateFakeValueForField($details['Type'], $col_name, $table_name, $details, $item);
            }
          }

          if (!empty($table_prime_keys[$table_name])) {
            $prime_key = $table_prime_keys[$table_name];

            if (!empty($item->{$prime_key}) && !empty($fields_and_values)) {
              $current_row_prime_key_value = $item->{$prime_key};

              try {
                $ice->database->update($table_name)
                  ->fields($fields_and_values)
                  ->condition($prime_key, $current_row_prime_key_value)
                  ->execute();
              }
              catch (\Exception $exception) {
                $ice->log(
                  dt(
                    'As this is WORK in progress  please check the method generateFakeValueForField  and --> this  @message', [
                      '@message' => $exception->getMessage(),
                    ]
                  ), 'error'
                );
              }
            }
          }
        }
      }
    }
  }
  // drupal_flush_all_caches();
  $ice->log(dt('Database has been cleaned'));
}
